/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _NAMESERVERCLIENT
#define _NAMESERVERCLIENT

/** \file NameServerClient.h
 * Definizione della classe fbfs::NameServerClient.
 * \ingroup NameServer
 */

#include <string>
#include <omniORB3/CORBA.h>

#include "../utils/URI.h"
#include "../utils/Prefs.h"
#include "../idl/Object.hh"
#include "../idl/Security.hh"

namespace fbfs {
	class NameServerClient;
};

/**
 * Ingloba il meccanismo di risoluzione dei nomi per individuare l'indirizzo di un oggetto.
 * Questa classe effettua i seguenti compiti:
 * - Dato uno steam di Prefs (preferenze), ricava da esso gli indirizzi per il NameServer
 *   primario e secondario.
 * - Risolve i nomi utilizzando di preferenza il NameServer primario; in caso esso non risponde
 *   prova con il NameServer secondario.
 * - Qualora un NameServer restituisca come indirizzo dell'oggetto un URI corrispondente ad
 *   un altro NameServer, utilizza esso per cercare di risolvere iterativamente l'indirizzo.
 * \ingroup NameServer
 */
class fbfs::NameServerClient {
public:
	/**
	 * Crea un nuovo NameServerClient utilizzando le preferenze contenute nello stream Prefs.
	 * Qualora non vi siano in Prefs i campi nameserver1 e nameserver2, attribuisce ad essi
	 * un valore di default, specificabile come parametri opzionali.
	 * \param o ORB da utilizzare per l'accesso agli agenti remoti.
	 * \param p stream di preferenze.
	 * \param primary indirizzo di default per il NameServer primario.
	 * \param secondary indirizzo di default per il NameServer secondario.
	 */
	NameServerClient(const CORBA::ORB_ptr &o, const utils::Prefs &p, const string &pri, const string &sec);

	/**
	 * Ricerca l'indirizzo dell'ObjectManager relativo all'oggetto con identificatore id.
	 * \param id identificatore dell'oggetto da ricercare
	 * \return indirizzo dell'ObjectManager che gestisce l'oggetto.
	 * \return "" qualora l'oggetto non sia possibile risolvere il nome.
	 */
	utils::URI lookup(const fbfs::Obj_id &id);
private:
	utils::URI lookup_int(const fbfs::Obj_id &id, utils::URI uri) const;
	const utils::URI primary;
	const utils::URI secondary;
	const CORBA::ORB_ptr &orb;
	Security cert;
};

#endif
