/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/** \file ObjectManager_i.cc
 * Implementazione dei metodi della classe fbfs::ObjectManager_i
 * \ingroup ObjectManager
 */
 
#include "ObjectManager_i.h"
#include "../utils/URI.h"
#include "../utils/NameServer.h"
#include "../utils/TimeStamp.h"
#include "../utils/Rand.h"
#include "../idl/Storage.hh"
#include "../Object/Properties.h"

#include <iostream.h>

using namespace fbfs;
using namespace utils;

fbfs::ObjectManager_i::ObjectManager_i(const utils::Prefs &p, const CORBA::ORB_ptr &o)
		: prefs(p), orb(o) {
		
		// Scandisco il file con l'elenco degli hosts e riempio la struttura servers.
		ifstream file;
		file.open(prefs.get("hostfile", "objectmanager-server.hosts").c_str());
		while(!file.eof()) {
			string s;
			file >> s;
			URI uri(s);
			if(uri.isValid())
				servers.push_front(s);
		};

		// Eseguo la funzione di convalida degli Storage e di preparazione della
		// struttura Repository.
		validate();

		// Ora rep contiene il repository aggiornato e consistente.
		// Posso attivare il servizio.
};

char *fbfs::ObjectManager_i::lookup(const Security &cert, const Obj_id &id)
		throw(NotFound, SecurityException, InternalError) {
	string name(id.name);
	lock.lock(name);
	
	if( rep.count(name) == 0 ) {
		lock.unlock(name);
		throw NotFound();
	};
	
	Servers &s = rep[name];

	if( s.size() < 1 ) {
		lock.unlock(name);
		throw InternalError();
	};

	// Sceglie tra i server quello pi opportuno. Potrebbe essere effettuato
	// secondo politiche di vicinanza, per esempio analizzando il certificato
	// di identit del richiedente.
	// Qui: scelta casuale tra i server disponibili.
	int n = utils::rand(s.size());
	Servers::iterator i = s.begin();
	while(n-->0)
		i++;
	lock.unlock(name);
	return strdup((*i).c_str());
};

void fbfs::ObjectManager_i::remove(const Security &cert, const Obj_id &id)
		throw(PartialExecution, NotFound, SecurityException, InternalError) {
	string name(id.name);
	lock.lock(name, true);
	
	if( rep.count(name) == 0 ) {
		lock.unlock(name);
		throw NotFound();
	};
	
	Servers &s = rep[name];

	if( s.size() < 1) {
		lock.unlock(name);
		throw InternalError();
	};

	// Devo contattare ogni server, cancellare l'oggetto da ogniuno di essi
	// ed alla fine rimuovere l'entry da rep.

	// Politica in caso di guasti: cerco di cancellare l'oggetto da ogni
	// Storage collegato. Se qualcuno d errore, proseguo fino alla fine
	// con gli altri Storage e ritorno InternalError().
	bool failed = false;
	
	for(Servers::iterator i = s.begin(); i != s.end(); i++) {
		URI uri(*i);
		try {
			CORBA::Object_var obj = utils::NameServer::get_object(orb, uri);
			fbfs::Storage_var server = fbfs::Storage::_narrow(obj);
			if(CORBA::is_nil(server))
				throw InternalError();

			server->remove(cert, id);
		} catch (...) {
			// Attenzione: in questo caso maledetto il repository pu essere in
			// uno stato inconsistente.
			failed = true;
		}
	}
	rep.erase(name);
	lock.unlock(name);
	
	if(failed)
		throw PartialExecution();
};

void fbfs::ObjectManager_i::put(const Security &cert, const Obj_id &id, const Obj &o, const Obj_prop &p)
		throw(PartialExecution, AlreadyPresent, SecurityException, InternalError) {

	string name(id.name);
	lock.lock(name, true);
	
	// Verifico che l'oggetto non sia gi nel repository
	if(rep.count(name) != 0) {
		lock.unlock(name);
		throw AlreadyPresent();
	};
		
	// Genero una struttura Obj_ts per l'oggetto.
	Obj_ts ts;
	ts.time = utils::TimeStamp::current();
	
	// Stabilisco da Obj_prop la politica di replicazione e la metto in atto.
	switch(p.type) {
		case Properties::DONT_SAVE :
			break;
		case Properties::REPLICATE_ON_CREAT :
		case Properties::REPLICATE_ALWAYS :
		case Properties::REPLICATE_MIN_MAX :
		{
			int ns = servers.size();
			int n = p.value_a;
			if(n>ns)
				n=ns;
			// itemserver sar l'elenco dei server che contengolo l'oggetto desiderato.
			Servers itemservers;
			
			// Metto gli oggetti nei server, scandendo in modo ciclico la lista.
			// Se arrivo in fondo, riparto dall'inizio.
			Servers::iterator i = servers.begin();
			for(int skip = utils::rand(servers.size()); skip-- > 0; i++);
			bool failed = false;
			while(n--) {
				const string &s = (*i);
				URI uri(s);
			
				// Contatto il server "s" e cerco di mettergli dento l'oggetto
				// desiderato.
				try{
					CORBA::Object_var obj = utils::NameServer::get_object(orb, uri);
					fbfs::Storage_var server = fbfs::Storage::_narrow(obj);
					if(CORBA::is_nil(server))
						throw NotFound();
					server->put(cert, o, id, p, ts);
					itemservers.push_front(s);
				} catch (...) {
					// Potrei ritentare su altri server, ma la cosa indurre
					// deadlock difficilmente gestibili.
					// Meglio rinunciare. 
					failed = true;
				};
				// Incrementa l'iterator, ricominciando da capo se la lista  finita.
				i++;
				if(i == servers.end() )
					i = servers.begin();
			};
			
			// Nessun oggetto messo nel repository
			if(itemservers.size()==0) {
				lock.unlock(name);
				throw InternalError();
			}
				
			// In questo caso la creazione non rispecchia le necessit imposte.
			// Cancello l'oggetto dagli Storage che ce l'hanno e ritorno InternalError.
			if(p.type == Properties::REPLICATE_MIN_MAX && itemservers.size() < p.value_b) {
				Servers::iterator i = itemservers.begin();
				while(i != itemservers.end()) {
					URI uri(*i);
					try {
						CORBA::Object_var obj = utils::NameServer::get_object(orb, uri);
						fbfs::Storage_var server = fbfs::Storage::_narrow(obj);
						if(CORBA::is_nil(server))
							throw InternalError();
						server->remove(cert, id);
					} catch (...) {};
				};
				lock.unlock(name);
				throw InternalError();
			};

			// Metto la coppia [oggetto . lista dei server] nel repository.
			rep[name] = itemservers;
			
			if(failed) {
				lock.unlock(name);
				throw PartialExecution();
			};
			break;
		};
		default:
			lock.unlock(name);
			throw InternalError();
	};
	lock.unlock(name);
};

void fbfs::ObjectManager_i::reparse(const Security &cert) throw(SecurityException) {
};
