/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ObjectManager_i.h"
#include "DataStructures.h"

#include "../idl/Storage.hh"
#include "../idl/Security.hh"
#include "../Object/Properties.h"
#include "../utils/URI.h"
#include "../utils/NameServer.h"
#include "../utils/Rand.h"

#include <values.h>

using namespace utils;

/** \file Validate.cc
 * Definizione della funzione fbfs::ObjectManager_i::validate()
 */
 
/**
 * Eseguito in single user mode prima dell'avvio del servizio si occupa di leggere
 * lo stato degli Storage che gestisce, di portarli in uno stato consistente e di
 * inizializzare la struttura Repository, che contiene una lista degli oggetti gestiti.
 * \ingroup ObjectManager
 */
void fbfs::ObjectManager_i::validate() {
	TempRepository tr;
	
	// PRIMA FASE: Query a tutti gli Storage registrati (nella struttura servers) del
	// loro contenuto.
	// Creazione della struttura TempRepository, mappa che associa ad ogni Id di oggetto
	// l'elenco delle terne <ServerOspitante, Obj_prop, Obj_ts> che rapprenentano ogni
	// copia dell'oggetto negli Storage.
	for(Servers::iterator it = servers.begin(); it != servers.end(); it++) {
		// Per ogni Storage faccio quanto segue:
		// - Lo contatto; se non riesco a raggiungerlo lo elimino dalla lista.
		// - Eseguo una get_dir e metto aggiorno la struttura tr.
		try {
			URI uri((*it).c_str());
			CORBA::Object_var obj = utils::NameServer::get_object(orb, uri);
			fbfs::Storage_var server = fbfs::Storage::_narrow(obj);
			if(CORBA::is_nil(server))
				throw NotFound();
			
			Obj_infoList *list = server->list(cert);
			for(int i=0; i<list->length(); i++) {
				Obj_info &info = (*list)[i];
				string name(info.id.name);
				InstanceInfo instance;
				instance.server = (*it);
				instance.prop = info.prop;
				instance.ts = info.ts;
				tr[name].push_front(instance);
			};
		} catch (...) {
			// In caso di errore --qualunque esso sia-- elimino il server dall'elenco.
			servers.erase(it);
		};
	};
	// Qui la struttura TempRepository  completa.

	// SECONDA FASE: Validazione della struttura TempRepository (con eventuali operazioni
	// sugli storage per riportare la consistenza) e creazione della struttura Repository.
	// Procedo come segue:
	// - Per ogni oggetto nel TempRepository, invoco una procedura che lo convalidi, occupandosi
	//   di rendere consistenti le istanze.
	// - Aggiungo ogni oggetto valido nella struttura Repository rep, con l'elenco dei server
	//   che lo forniscono.

	rep = Repository();
	for( TempRepository::iterator i = tr.begin(); i != tr.end(); i++ ) {
		Servers s = validate( (*i).first, (*i).second );
		if(s.size() != 0)	
			rep[(*i).first] = s;
	};
};

/*
Algoritmo per la verifica e la eventuale ricostituzione della consistenza:

Data una coppia (nome oggetto, lista<istanze>), eseguo:

- Individuo tra le istanze quella con il ts pi recente. Se vi sono istanze con
ts(istanza) < ts(istanza_con_ts_pi_recente), le elimino in quanto esse rappresentano
copie vecchie di oggetti (le elimino anche dagli Storage dove sono contenute).

- A questo punto Obj_id e Obj_ts sono sistemate. Bisogna individuare il tipo di propriet
che l'oggetto deve avere. Per fare questo agisco secondo le seguenti regole di sostituzione:

DON_SAVE -> REPLICATE_ON_CREAT(value_a=0);
REPLICATE_ON_CREAT -> REPLICATE_ALWAYS(value_a=value_a.old);
REPLICATE_ALWAYS -> REPLICATE_MIN_MAX(value_a=value_b=value_a.old);

Applico solamente le regole di promozione necessarie affinch tutte le istanze abbiano il
medesimo insieme di propriet.
Queste propriet saranno la struttura Obj_prop dell'Oggetto definitivo nel repository.

- Per tutte le istanze con Obj_prop non corretto, le sistemo.

- Verifico che l'insieme di istanze verifichi globalmente i limiti imposti dalle propriet,
creando (o distruggendo, ma eventualit molto rara) ulteriori copie.
*/

// Cancella un oggetto da un dato server, ignorando gli errori.
static void remove(const string &server, const string &objname, CORBA::ORB_ptr &orb, const fbfs::Security &cert) {
	try {
		URI uri(server.c_str());
		CORBA::Object_var obj = utils::NameServer::get_object(orb, uri);
		fbfs::Storage_var server = fbfs::Storage::_narrow(obj);
		if(!CORBA::is_nil(server)) {
			fbfs::Obj_id id;
			id.name = CORBA::string_dup(objname.c_str());
			server->remove(cert, id);
		};
	} catch (...) {};
};

fbfs::Servers fbfs::ObjectManager_i::validate(const string &objname, list<InstanceInfo> &infos) {
	int ts = MININT;
	list<InstanceInfo> newinfos = list<InstanceInfo>();
	list<InstanceInfo>::iterator it;
	for(it = infos.begin(); it != infos.end(); it++) {
		if((*it).ts.time == ts) {
			// Bene, aggiungo l'istanza alla lista delle istanze valide
			newinfos.push_front(*it);
		} else if((*it).ts.time > ts) {
			// L'istanza  pi recente di quelle in lista, che devono essere quindi
			// scartate e inserisco nella lista la nuova istanza.
			for(list<InstanceInfo>::iterator i = newinfos.begin(); i != newinfos.end(); i++)
				::remove((*i).server, objname, orb, cert);
			newinfos = list<InstanceInfo>();
			newinfos.push_front(*it);
			ts = (*it).ts.time;
		} else {
			// se invece i.ts.time < ts allora devo scartare l'istanza corrente
			::remove((*it).server, objname, orb, cert);
		};
	};		

	// Qui ho eliminato dal repository le istanze "vecchie" (con ts diverso dal
	// ts dell'istanza pi recente) e in newinfos ho l'elenco delle istanze valide.
	
	// Determino le propriet, estrapolandole dalle propriet delle istanze.
	Obj_prop np;
	np.value_a = 0;
	np.value_b = 0;
	
	for(it = newinfos.begin(); it != newinfos.end(); it++) {
		Obj_prop &props = (*it).prop;
		switch (props.type) {
			case Properties::REPLICATE_MIN_MAX :
				if(props.value_b > np.value_b)
					np.value_b = props.value_b;
				if(props.value_a > np.value_a)
					np.value_a = props.value_a;
				break;
			case Properties::REPLICATE_ALWAYS :
				if(props.value_a > np.value_b)
					np.value_b = props.value_a;
			case Properties::REPLICATE_ON_CREAT :
				if(props.value_a > np.value_a)
					np.value_a = props.value_a;
			case Properties::DONT_SAVE :
				;
		};		
	};

	if(np.value_b == 0) {
		np.type = Properties::REPLICATE_ON_CREAT;
		if(np.value_a == 0)
			np.type = Properties::DONT_SAVE;
	} else if(np.value_a == np.value_b) {
		np.type = Properties::REPLICATE_ALWAYS;
	} else {
		np.type = Properties::REPLICATE_MIN_MAX;
	};

	// Ora mi occupo di garantire che il gruppo di istanze soddisfi le propriet.
	// In particolare deve essere (numero istanze) >= min( np.value_a, (numero di server) ).
	// Se non sono abbastanza, creo copie delle istanze.

	// Seleziono i server in cui mettere le nuove copie:
	Servers tempservers = servers;	// Servers disponibili per metterci copie
	Servers newservers;				// Servers gi utilizzati
	for( it = newinfos.begin(); it != newinfos.end(); it++) {
		tempservers.remove( (*it).server );
		newservers.push_front( (*it).server );
	};
	
	if(newinfos.size() >= np.value_b || newinfos.size() == servers.size() ) {
		// OK. Tutto aposto gi.
		return newservers;
	};

	// Devo creare le istanze rimanenti e metterle su dei server che non le contengano
	// gi.

	if(np.value_a > servers.size() )
		np.value_a = servers.size();
		
	// Numero di copie da creare
	int count = np.value_a - newinfos.size();

	Servers::iterator vecchio = newservers.begin();
	Servers::iterator nuovo = tempservers.begin();
	
	Servers finalservers = newservers;
	
	// Salto un numero a caso di server in vecchio, cos che gli oggetti siano
	// disposti uniformemente nei Servers.
	for( int n = utils::rand(newservers.size()); n--; nuovo++);
	
	for(int maxcycles=2*servers.size() ; count && maxcycles  ; nuovo++, vecchio++, count--, maxcycles--) {
		if( vecchio == newservers.end() )
			vecchio = newservers.begin();
		if( nuovo == tempservers.end() )
			nuovo = tempservers.begin();

		// Eseguo una copia utilizzando il metodo put_from_peer() del
		// nuovo storage dal vecchio storage.
		try {
			URI uri(*nuovo);
			CORBA::Object_var obj = utils::NameServer::get_object(orb, uri);
			fbfs::Storage_var server = fbfs::Storage::_narrow(obj);
			if(!CORBA::is_nil(server)) {
				Obj_id id;
				id.name = CORBA::string_dup(objname.c_str());
				server->put_from_peer(cert, id, (*vecchio).c_str() );
				finalservers.push_front(*nuovo);
			};
		} catch (...) {
			// Se c' qualche errore, me ne sbatto e provo su un server nuovo.
			count++;
		};
	};
	
	return finalservers;
};	
