/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/** \file Proxy_i.cc
 * Implementazione dei metodi della classe fbfs::Proxy_i.
 * \ingroup Proxy
 */
	
#include "Proxy_i.h"
#include "../idl/Storage.hh"
#include "../idl/Object.hh"
#include "../idl/ObjectManager.hh"
#include "../Object/Object.h"
#include "../utils/File.h"
#include "../utils/NameServer.h"
#include "../utils/URI.h"

using utils::File;

fbfs::Proxy_i::Proxy_i(utils::Prefs &p, const CORBA::ORB_ptr &o)
	: prefs(p), orb(o), dir(p.get("cachedir","data")), nsclient(o, p, "", "") {};

fbfs::Obj *fbfs::Proxy_i::get(const Security &cert, const Obj_id &id)
		throw(NotFound, SecurityException, InternalError) {
	fbfs::Storage_var storage;
	string storage_name;
	Obj_ts rem_ts;
	string ident(id.name);

	// Accesso shared, in principio. Se poi non c' il file e devo
	// prelevarlo dal server, allora ottengo un accesso exclusive
	lock.lock(ident);
	
	int level=0;
	try {
		string strobj = File::read(ident, dir);
		level = 1;
		
		string strobj_ts = File::read(ident+".ts", dir );
		Obj_ts loc_ts = Object::string2Obj_ts(strobj_ts);
		level = 2;
		
		storage_name = File::read(ident+".server", dir);
		level = 3;
		
		utils::URI uri(storage_name);
		
		CORBA::Object_ptr obj = utils::NameServer::get_object(orb, uri);
		storage = fbfs::Storage::_narrow(obj);
		if(CORBA::is_nil(storage)) {
			lock.unlock(ident);
			throw InternalError();
		}
		level = 4;
		
		rem_ts = storage->get_ts(cert, id);
		
		if( Object::cmp_ts(loc_ts, rem_ts) ) {
			lock.unlock(ident);
			return Object::string2Obj(strobj);
		}
	} catch (...) {};

	// Da qui accesso esclusivo al file
	lock.unlock(ident);
	lock.lock(ident, true);
	
	// In ogni altro caso: file non trovato sul server remoto, errore del server
	// remoto, differenza di ts, etc, rimuovo l'oggetto dalla cache e tento di
	// prelevarlo dal server remoto, aggiornando la cache.
	try {
		switch(level) {
			case 4:
			case 3:
				File::remove(ident+".server", dir);
			case 2:
				File::remove(ident+".ts", dir);
			case 1:
				File::remove(ident, dir);
		}
	} catch (...) {
		lock.unlock(ident);
		throw InternalError();
	};

	// In caso di mancanza di file, fino ad ora non  ancora stato
	// contattato il server remoto. Devo risolvere il nome
	// dell'ObjectManager che gestisce l'oggetto e ricercare in esso
	// l'indirizzo di uno Storage.
				
	try {
		if(level==0) {
			utils::URI uri(nsclient.lookup(id));
			if(!uri.isValid())
				throw NotFound();
					
			CORBA::Object_ptr obj = utils::NameServer::get_object(orb, uri);
			fbfs::ObjectManager_var om = fbfs::ObjectManager::_narrow(obj);
			if(CORBA::is_nil(om))
				throw NotFound();
			
			// Qui ricerco l'indirizzo del server nell'Object Manager 
			try {
				CORBA::String_var s = om->lookup(cert, id);
				storage_name = string(s);
			
				CORBA::Object_ptr obj = utils::NameServer::get_object( orb, utils::URI(storage_name) );
				storage = fbfs::Storage::_narrow(obj);
				if(CORBA::is_nil(storage))
					throw InternalError();

				rem_ts = storage->get_ts(cert, id);
			} catch(NotFound) {
				throw NotFound();
			} catch (...) {
				throw InternalError();
			};
		};
		// Ora il server  contattato. Qualora non abbia prima ricevuto gi
		// rem_ts, lo leggo ora.
		if(level!=4)
			rem_ts = storage->get_ts(cert, id);

		// Prelevo il file.
		Obj *o = storage->get(cert, id);

		// Decido se metterlo in cache

		// Lo metto in cache, se ci riesco, senza sollevare polveroni in caso
		// di mancanza di riuscita.
		level = 0;
		try {
			File::write( Object::Obj2string(*o), ident, dir, true);
			level = 1;
			File::write( Object::Obj_ts2string(rem_ts), ident+".ts", dir, true);
			level = 2;
			File::write( storage_name, ident+".server", dir, true);
			level = 3;
		} catch (...) {
			// Se ci sono problemi, cerco di rimuovere i file rimasti.
			try{
				switch(level) {
					case 3:
						File::remove( ident+".server", dir );
					case 2:
						File::remove( ident+".ts", dir );
					case 1:
						File::remove( ident, dir );
				}
			} catch (...) {};
		};
		
		// Restituisco un riferimento all'oggetto.
		// Si noti che qualora si decida di introdurre delle eccezioni che possano
		// capitare tra il o = storage-get() e qui, bisogna provvedere alla rimozione
		// dell'area di memoria associata all'oggetto o.
		lock.unlock(ident);
		return o;
	} catch (NotFound) {
		lock.unlock(ident);
		throw NotFound();
	} catch (SecurityException) {
		lock.unlock(ident);
		throw SecurityException();
	} catch (...) {
		lock.unlock(ident);
		throw InternalError();
	};
};

void fbfs::Proxy_i::put(const Security &cert, const Obj_id &id, const Obj &obj, const Obj_prop &props)
		throw(AlreadyPresent, NotFound, SecurityException, InternalError) {
	// - Cerco di mettere l'oggetto nel repository.
	// - Se ci riesco aggiorno la cache col nuovo oggetto
	//	 (eventualmente cancellando l'oggetto precedente),
	//   prelevando il suo Obj_ts dall'ObjectManager opportuno.
	// - Altrimenti lascio la cache come prima.

	utils::URI uri(nsclient.lookup(id));
	if(!uri.isValid())
		throw NotFound();

	fbfs::ObjectManager_var om;
	try {
		CORBA::Object_ptr obj = utils::NameServer::get_object(orb, uri);
		om = fbfs::ObjectManager::_narrow(obj);
	} catch (...) { throw NotFound(); };
	if(CORBA::is_nil(om))
		throw NotFound();

	// Qui c' il riferimento all'ObjectManager.
	om->put(cert, id, obj, props);

	// Contattando l'ObjectManager, ottengo l'indirizzo di uno Storage,
	// che metto in storage_name. Poi contattando lo Storage recupero
	// la struttura Obj_ts.
	string ident(id.name);
	Obj_ts rem_ts; 

	string storage_name;
	try {
		CORBA::String_var s = om->lookup(cert, id);
		storage_name = string(s);
		
		CORBA::Object_ptr obj = utils::NameServer::get_object( orb, utils::URI(storage_name) );
		fbfs::Storage_var storage = fbfs::Storage::_narrow(obj);
		if(CORBA::is_nil(storage))
			throw InternalError();

		rem_ts = storage->get_ts(cert, id);
	} catch (...) {
		// Qui andrebbe _molto_ codice per capire cosa fare in questa eventualit
		// rara. Facciamo finta di niente:
		// "Never test for an error condition you don't know how to handle."
		return;
	};

	// Ottengo un accesso esclusivo all'oggetto.
	lock.lock(ident, true);
	
	// Lo metto in cache, se ci riesco, senza sollevare polveroni in caso
	// di mancanza di riuscita.
	int level = 0;
	try {
		File::write( Object::Obj2string(obj), ident, dir, true);
		level = 1;
		File::write( Object::Obj_ts2string(rem_ts), ident+".ts", dir, true);
		level = 2;
		File::write( storage_name, ident+".server", dir, true);
		level = 3;
	} catch (...) {
		// Se ci sono problemi, cerco di rimuovere i file rimasti.
		try{
			switch(level) {
				case 3:
					File::remove( ident+".server", dir );
				case 2:
					File::remove( ident+".ts", dir );
				case 1:
					File::remove( ident, dir );
			}
		} catch (...) {};
	};
	lock.unlock(ident);
};

void fbfs::Proxy_i::remove(const Security &cert, const Obj_id &id)
		throw(NotFound, SecurityException, InternalError) {
	// - Cerco di rimuovere l'oggetto dal repository.
	// - Se ci riesco cancello l'eventuale copia in cache.

	utils::URI uri(nsclient.lookup(id));
	if(!uri.isValid())
		throw NotFound();

	string ident(id.name);

	CORBA::Object_ptr obj = utils::NameServer::get_object(orb, uri);
	fbfs::ObjectManager_var om = fbfs::ObjectManager::_narrow(obj);
	if(CORBA::is_nil(om))
		throw NotFound();

	lock.lock(ident, true);
	try {
		om->remove(cert, id);
	} catch (...) {
		lock.unlock(ident);
		throw;
	};

	// Se il flusso arriva qui, l'oggetto esisteva nel repository remoto
	// ed  stato rimosso con successo.
	// Ora lo rimuovo dalla cache, se c'.
	// Approccio ottimista qualora non riesca a rimuoverlo. Anche se rimane,
	// infatti non fa danno.
	try { File::remove(ident, dir); } catch (...) {};
	try { File::remove(ident+".ts", dir); } catch (...) {};
	try { File::remove(ident+".server", dir); } catch (...) {};
	lock.unlock(ident);
};
