/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _fbfs_PROXY
#define _fbfs_PROXY 1

#include "../idl/Proxy.hh"
#include "../utils/Prefs.h"
#include "../utils/ShareLock.h"
#include "../NameServer/NameServerClient.h"

/** \file Proxy_i.h
 * Definizione della classe fbfs::Proxy_i.
 * Proxy_i  l'implementazione della interfaccia Proxy, come oggetto CORBA.
 * \ingroup Proxy
 */

namespace fbfs {
	class Proxy_i;
};

/**
 * FrontEnd di accesso a tutto il sistema distribuito.
 * Questa classe si occupa di inviare le richieste desiderate agli appositi ObjectManager;
 * di prelevare gli Obj deisderati dagli appositi Storage; di risolvere i nomi degli agenti
 * attraverso gli appositi NameServer.
 *
 * Essa mantiene una cache locale degli oggetti desidetati in lettura. La coerenza della
 * cache con il repository viene valutata ogni lettura, confrontanto la struttura Obj_ts
 * dell'oggetto locale con la corrispondente struttura dell'oggetto nel repository.
 * \ingroup Proxy
 */
class fbfs::Proxy_i : public POA_fbfs::Proxy, public PortableServer::RefCountServantBase {
public:
	/**
	 * Invocato al momento della creazione dell'oggetto sigleton utilizzato dal server
	 * CORBA. Qui  inizializzato lo stato del server.
	 */
	Proxy_i(utils::Prefs &prefs, const CORBA::ORB_ptr &orb);

	/**
	 * Invocato alla chiusura del server.
	 */
	virtual ~Proxy_i() {};

	/**
	 * Preleva un oggetto dal repository.
	 * Consente di reperire un'oggetto nel repository, risalendo alla sua locazione
	 * attraverso i NameServer (i cui indirizzi sono specificati nel file di configurazione
	 * letto dal server alla partenza).
	 *
	 * Qualora sia presente un oggetto nella cache locale,
	 * col voluto Obj_id, viene valutata la consistenza dell'oggetto locale con quello nel
	 * repository confrontandone le strutture Obj_ts. Se esse coincidono, viene restituita
	 * copia dell'oggetto locale. Altrimenti l'oggetto locale viene rimosso e si procede
	 * alla lettura dallo storage remoto ed all'inserimento dell'oggetto remoto nella cache
	 * locale.
	 *
	 * \param cert certificato di sicurezza del richiedente.
	 * \param id identificativo dell'oggetto richiesto.
	 * \return riferimento ad una copia locale dell'oggetto richiesto.
	 * \exception NotFound l'oggetto richiesto non  raggiungibile nel repository
	 * \exception SecurityException il richiedente non ha i diritti per ottenere il riferimento desiderato.
	 * \exception InternalError il server ha commesso un errore interno nel processare la richiesta.
	 */ 
	 fbfs::Obj *get(const Security &cert, const Obj_id &id)
			throw(NotFound, SecurityException, InternalError);

	/**
	 * Inserisce un oggetto nel repository.
	 * L'inserimento consta delle seguenti fasi:
	 * - Verifica della presenza remota di un file con il medesimo Obj_id (eventualmente
	 *   consultando anche la cache locale.
	 * - In caso negativo inserimento (tramite l'ObjectManager opportuno) dell'oggetto
	 *   nel repository.
	 * - Inserimento dell'oggetto nella cache locale.
	 * - Copia del Obj_ts dall'oggetto nel repository all'oggetto locale.
	 *
	 * \param cert certificato di sicurezza del richiedente.
	 * \param id identificativo dell'oggetto da inserire.
	 * \param obj oggetto da inserire.
	 * \param props propriet dell'oggetto da inserire.
	 * \exception AlreadyPresent nel repository  gi presente un oggetto col medesimo Obj_id.
	 * \exception NotFound impossibile risalire all'ObjectManager adatto a gestire l'oggetto.
	 * \exception SecurityException il richiedente non ha i diritti per inserire l'oggetto nel repository.
	 * \exception InternalError il server ha commesso un errore interno nel processare la richiesta.
	 */
	void put(const Security &cert, const Obj_id &id, const Obj &obj, const Obj_prop &props)
			throw(AlreadyPresent, NotFound, SecurityException, InternalError);

	/**
	 * Rimuove un oggetto dal repository.
	 * La rimozione consta delle seguenti fasi:
	 * - Identificazione del ObjectManager responsabile dell'eventuale oggetto identificato da id
	 *   (anche attraverso query nella cache locale).
	 * - Richiesta di rimozione dell'oggetto dal repository presso l'ObjectManager opportuno.
	 * - Rimozione della eventuale copia locale dell'oggetto,
	 * 
	 * \param cert certificato di sicurezza del richiedente.
	 * \param id identificativo dell'oggetto da cancellare.
	 * \exception NotFound l'oggetto richiesto non  raggiungibile nel repository
	 * \exception SecurityException il richiedente non ha i diritti per cancellare l'oggetto.
	 * \exception InternalError il server ha commesso un errore interno nel processare la richiesta.
	 */
	void remove(const Security &cert, const Obj_id &id)
			throw(NotFound, SecurityException, InternalError);

private:
	utils::Prefs &prefs;
	const string dir;
	CORBA::ORB_ptr orb;
	NameServerClient nsclient;
	utils::ShareLock lock;
};

/** \example Proxyget.cc
 * Legge un file dal repository.
 */

/** \example Proxyput.cc
 * Inserisce un file nel repository.
 */

/** \example Proxyremove.cc
 * Rimuove un file dal repository.
 */

#endif
