/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/** \file Storage_i.cc
 * Implementazione dei metodi della classe fbfs::Storage_i
 * \ingroup Storage
 */
 
#include "Storage_i.h"

#include "../utils/File.h"
#include "../utils/NameServer.h"
#include "../Object/Object.h"
#include "../idl/Storage.hh"

#include <dirent.h>

Storage_i::Storage_i(Prefs &p, const CORBA::ORB_ptr &o) : prefs(p), orb(o) {;
	dirname = prefs.get("datadir", "data");
	metadirname = prefs.get("metadatadir", dirname);
};
	
Obj *Storage_i::get(const Security &cert, const Obj_id &id)
					throw(NotFound, SecurityException, InternalError) {
	string ident(id.name);
	try {
		lock.lock(ident);
			string strobj = File::read( ident, dirname );
		lock.unlock(ident);
		return Object::string2Obj(strobj);
	} catch (File::NotFound) {
		lock.unlock(ident);
		throw NotFound();
	} catch (...) {
		lock.unlock(ident);
		throw InternalError();
	};
};

Obj_ts Storage_i::get_ts(const Security &cert, const Obj_id &id)
					throw(NotFound, SecurityException, InternalError) {
	string ident(id.name);
	try {
		lock.lock(ident);
			string strobj = File::read( ident + ".ts", metadirname );
		lock.unlock(ident);
		return Object::string2Obj_ts(strobj);
	} catch (File::NotFound) {
		lock.unlock(ident);
		throw NotFound();
	} catch (...) {
		lock.unlock(ident);
		throw InternalError();
	};
};

Obj_prop Storage_i::get_prop(const Security &cert, const Obj_id &id)
					throw(NotFound, SecurityException, InternalError) {
	string ident(id.name);
	try {
		lock.lock(ident);
			string strobj = File::read( ident + ".prop", metadirname );
		lock.unlock(ident);
		return Object::string2Obj_prop(strobj);
	} catch (File::NotFound) {
		lock.unlock(ident);
		throw NotFound();
	} catch (...) {
		lock.unlock(ident);
		throw InternalError();
	};
};

void Storage_i::put(const Security &cert, const Obj &obj, const Obj_id &id, const Obj_prop &prop,
					const Obj_ts &ts) throw(AlreadyPresent, SecurityException, InternalError) {
	string ident(id.name);
	string obj_str = Object::Obj2string(obj);
	string obj_ts_str = Object::Obj_ts2string(ts);
	string obj_prop_str = Object::Obj_prop2string(prop);

	lock.lock(ident, true);
	
	// Cerca di scrivere il file. Se c' gi, esci con throw AlreadyPresent().
	try {
		File::write(obj_str, ident, dirname);
	} catch (File::AlreadyPresent) {
		lock.unlock(ident);
		throw AlreadyPresent();
	};

	try {
		// Cerca di scrivere i file .ts e .prop cancellando eventualmente le vecchie versioni.
		File::write(obj_ts_str, ident+".ts", metadirname, true);
		File::write(obj_prop_str, ident+".prop", metadirname, true);
	} catch (...) {
		// Se durante l'operazione si verifica un errore, cerca di cancellare i file prodotti
		// e esci con throw InternalError().
		try { File::remove(ident, dirname); } catch (...) {};
		try { File::remove(ident+".ts", metadirname); } catch (...) {};
		try { File::remove(ident+".prop", metadirname); } catch (...) {};
		lock.unlock(ident);
		throw InternalError();
	};
	lock.unlock(ident);
};

Obj_infoList *Storage_i::list(const Security &cert) throw(SecurityException, InternalError) {
	DIR *dir;
	struct dirent * entry;
	int count=0;
	
	dir = opendir( metadirname.c_str() );
	if(dir == NULL)
		throw InternalError();
	while( entry=readdir(dir) ) {
		string name(entry->d_name);
		if(name.find(".ts")==name.length()-3 && name.length()>3 )
			count++;
	};
	free(entry);
	closedir(dir);
	
	Obj_infoList *list = new Obj_infoList(count);
	list->length(count);

	dir = opendir( metadirname.c_str() );
	if(dir == NULL)
		throw InternalError();
	int i=0;
	while( entry=readdir(dir) ) {
		string name(entry->d_name);
		if(name.find(".ts")==name.length()-3 && name.length()>3 ) {
			name = name.substr(0,name.length()-3);
			
			Obj_info info;
			info.id.name= CORBA::string_dup(name.c_str());
			try{ info.ts = Object::string2Obj_ts(File::read(name+".ts", metadirname)); } catch (...) {};
			try{ info.prop = Object::string2Obj_prop(File::read(name+".prop", metadirname)); } catch (...) {};
			
			(*list)[i++] = info;
		};
	};
	free(entry);
	closedir(dir);
	return list;
};

void Storage_i::put_from_peer(const Security &cert, const Obj_id &id, const char *peer)
					throw(NotFound, PeerNotAvailable, AlreadyPresent, SecurityException, InternalError) {

	// Accesso esclusivo all'oggetto
	string ident(id.name);
	lock.lock(ident, true);
	
	// Indirizzo del server da prelevare
	utils::URI uri(peer);

	Storage_var storage;
	
	try {
		// Cerca di ottenere un riferimento al server
		CORBA::Object_var stor_obj = utils::NameServer::get_object(orb, uri);
		storage = fbfs::Storage::_narrow(stor_obj);
	} catch(...) {
		lock.unlock(ident);
		throw PeerNotAvailable();
	};
	
	Obj *obj;
	Obj_ts obj_ts;
	Obj_prop obj_prop;
	try {
		obj = storage->get(cert, id);
		obj_ts = storage->get_ts(cert, id);
		obj_prop = storage->get_prop(cert, id);
	} catch(NotFound) {
		lock.unlock(ident);
		throw NotFound();
	} catch(SecurityException) {
		lock.unlock(ident);
		throw SecurityException();
	} catch(...) {
		lock.unlock(ident);
		throw PeerNotAvailable();
	};
	
	try{
		lock.unlock(ident);
		put(cert, *obj, id, obj_prop, obj_ts);
	} catch(AlreadyPresent) {
		throw AlreadyPresent();
	} catch(SecurityException) {
		throw SecurityException();
	} catch(...) {
		throw InternalError();
	};
};

void Storage_i::remove(const Security &cert, const Obj_id &id)
					throw(NotFound, SecurityException, InternalError) {
	string ident(id.name);

	// Accesso esclusivo all'oggetto
	lock.lock(ident, true);
	
	bool failed = false;
	try {
		File::remove(ident, dirname);
	} catch (File::NotFound) {
		lock.unlock(ident);
		throw NotFound();
	} catch (...) {
		failed = true;
	};
	try { File::remove(ident+".ts", metadirname); } catch (NotFound) {} catch (...) { failed = true;};
	try { File::remove(ident+".prop", metadirname); } catch (NotFound) {} catch (...) { failed = true;};

	lock.unlock(ident);
	if(failed)
		throw InternalError();
};
