/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _fbfs_STORAGE
#define _fbfs_STORAGE 1

#include <omniORB3/CORBA.h>

#include "../idl/Storage.hh"

#include "../utils/Prefs.h"
#include "../utils/ShareLock.h"

/** \file Storage_i.h
 * Definizione della classe fbfs::Storage_i.
 * Storage_i  l'implementazione della interfaccia Storage, come oggetto CORBA.
 * \ingroup Storage
 */

namespace fbfs {
		class Storage_i;
};

using namespace fbfs;
using namespace utils;

/**
 * Si occupa di mantenere un repository di oggetti di tipo Obj.
 * Implementazione di un Server in grado di mantenere in forma persistente
 * delle strutture di tipo Obj, ogniuna corredata di una struttura ausiliaria
 * fbfs::Obj_prop, che definisce le propriet di persistenza dell'oggetto. Esse sono
 * utilizzate dal server FileManager, ma sono mantenute in forma persistente assieme
 * all'oggetto a cui si riferiscono per permettere il recovery in caso di crash del
 * server FileManager.
 * 
 * Ogni oggetto  inoltre corredato da una struttura di tipo Obj_ts che garantisce
 * che la coppia (fbfs::Obj_id, fbfs::Obj_ts) individui in modo univoco un oggetto nel sistema:
 * ci permette di effettuare il cacheing di oggetti presso un server remoto e di
 * verificare la consistenza delle copie fredde con quelle calde con la sola verifica
 * dell'uguaglianza di fbfs::Obj_id e fbfs::Obj_ts. La struttura fbfs::Obj_ts  generata dal FileManager,
 * affinch tutte le copie attive abbiano la medesima fbfs::Obj_ts.
 * \ingroup Storage
 */
class fbfs::Storage_i : public POA_fbfs::Storage, public PortableServer::RefCountServantBase {
public:
	/**
	 * Invocato al momento della creazione dell'oggetto sigleton utilizzato dal server
	 * CORBA. Qui  inizializzato lo stato del server.
	 */
	Storage_i(Prefs &prefs, const CORBA::ORB_ptr &orb);

	/**
	 * Invocato alla chiusura del server.
	 */
	virtual ~Storage_i() {};
	
	/**
	 * Legge dal repository un oggetto, dato il suo fbfs::Obj_id.
	 * \param cert certificato di sicurezza del richiedente.
	 * \param id identificativo dell'oggetto.
	 * \return oggetto identificato dall'fbfs::Obj_id id.
	 * \exception NotFound nel repository non  presente l'oggetto indicato da fbfs::Obj_id.
	 * \exception SecurityException il richiedente non ha i diritti per ottenere l'oggetto desiderato.
	 * \exception InternalError il server non ha commesso un errore interno nel processare la richiesta.
	 */
	Obj *get(const Security &cert, const Obj_id &id)
			throw(NotFound, SecurityException, InternalError);

	/**
	 * Ottiene la struttura fbfs::Obj_ts associata ad un oggetto.
	 * \param cert certificato di sicurezza del richiedente.
	 * \param id identificativo dell'oggetto.
	 * \return struttura fbfs::Obj_ts associata all'oggetto identificato dall'fbfs::Obj_id id.
	 * \exception NotFound nel repository non  presente l'oggetto indicato da fbfs::Obj_id.
	 * \exception SecurityException il richiedente non ha i diritti per ottenere l'oggetto desiderato.
	 * \exception InternalError il server non ha commesso un errore interno nel processare la richiesta.
	 */
	Obj_ts get_ts(const Security &cert, const Obj_id &id)
			throw(NotFound, SecurityException, InternalError);

	/**
	 * Ottiene la struttura Obj_prop associata ad un oggetto.
	 * \param cert certificato di sicurezza del richiedente.
	 * \param id identificativo dell'oggetto.
	 * \return struttura fbfs::Obj_prop associata all'oggetto identificato dall'fbfs::Obj_id id.
	 * \exception NotFound nel repository non  presente l'oggetto indicato da fbfs::Obj_id.
	 * \exception SecurityException il richiedente non ha i diritti per ottenere l'oggetto desiderato.
	 * \exception InternalError il server non ha commesso un errore interno nel processare la richiesta.
	 */
	Obj_prop get_prop(const Security &cert, const Obj_id &id)
			throw(NotFound, SecurityException, InternalError);

	/**
	 * Inserisce un oggetto nel repository.
	 * \param cert certificato di sicurezza del richiedente.
	 * \param obj oggetto da inserire nel repository.
	 * \param id identificativo dell'oggetto.
	 * \param prop struttura fbfs::Obj_prop associata all'oggetto.
	 * \param ts struttura fbfs::Obj_ts associata all'oggetto.
	 * \exception AlreadyPredent nel repository  gi presente un oggetto con fbfs::Obj_id uguale a id.
	 * \exception SecurityException il richiedente non ha i diritti per inserire l'oggetto nel repository.
	 * \exception InternalError il server non ha commesso un errore interno nel processare la richiesta.
	 */
	void put(const Security &cert,const Obj &obj,const Obj_id &id,const Obj_prop &prop,const Obj_ts &ts)
			throw(AlreadyPresent, SecurityException, InternalError);

	/**
	 * Ottiene una lista di identificatori degli oggetti contenuti nel repository.
	 * \param cert certificato di sicurezza del richiedente.
	 * \return lista di fbfs::Obj_info.
	 * \exception SecurityException il richiedente non ha i diritti per ottenere la lista.
	 * \exception InternalError il server non ha commesso un errore interno nel processare la richiesta.
	 */
	Obj_infoList *list(const Security &cert)
			throw(SecurityException, InternalError);

	/**
	 * Inserisce nel repository un oggetto prelevato da un altro Storage server.
	 * \param cert certificato di sicurezza del richiedente.
	 * \param id identificativo dell'oggetto.
	 * \param peer URI dello Storage server da cui prelevare l'oggetto.
	 * \exception NotFound sul server Storage remoto non  presente l'oggetto desiderato.
	 * \exception PeerNotAvailable il server remoto non  raggiungibile.
	 * \exception AlreadyPresent nel repository  gi presente un oggetto con fbfs::Obj_id uguale a id.
	 * \exception SecurityException il richiedente non ha i diritti per eseguire l'operazione.
	 * \exception InternalError il server non ha commesso un errore interno nel processare la richiesta.
	 */
	void put_from_peer(const Security &cert, const Obj_id &id, const char *peer)
			throw(NotFound, PeerNotAvailable, AlreadyPresent, SecurityException, InternalError);

	/**
	 * Rimuove un oggetto dal repository.
	 * \param cert certificato di sicurezza del richiedente.
	 * \param id identificativo dell'oggetto.
	 * \exception NotFount nel repository non  presente l'oggetto indicato da fbfs::Obj_id.
	 * \exception SecurityException il richiedente non ha i diritti per eseguire l'operazione.
	 * \exception InternalError il server non ha commesso un errore interno nel processare la richiesta.
	 */
	void remove(const Security &cert, const Obj_id &id)
			throw(NotFound, SecurityException, InternalError);

private:
	Prefs &prefs;
	CORBA::ORB_ptr orb;
	string dirname;
	string metadirname;
	utils::ShareLock lock;
};

/** \example get.cc
 * Legge un file dallo Storage.
 */

/** \example get_prop.cc
 * Legge le propriet di un file dallo Storage.
 */

/** \example get_ts.cc
 * Legge il timestamp di un file dallo Storage.
 */

/** \example put.cc
 * Introduce un file in uno Storage.
 */

/** \example put_from_peer.cc
 * Copia un file in uno Storage da un altro Storage.
 */

/** \example remove.cc
 * Rimuove un file da uno Storage.
 */

#endif
