/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _utils_LOCK
#define _utils_LOCK

/** \file Lock.h
 * Definizione della classe utils::Lock.
 * \ingroup Utils
 */

#include <set>
#include <string>
#include "String.h"

namespace utils {
	class Lock;
};

/**
 * Consente l'accesso esclusivo ad una risorsa ad un gruppo di thread.
 * Dato un insieme di risorse critiche identificate ogniuna con una string, consente
 * di garantire l'accesso ad ogniuna di esse in modo esclusivo ad un singolo thread
 * di esecuzione.
 * \ingroup Utils
 */
class utils::Lock {
	public:
		/**
		 * Crea un oggetto di tipo Lock, allocando le risorse necessarie al suo utilizzo.
		 */
		Lock();

		/**
		 * Distrugge un oggetto di tipo Lock, liberando le risorse utilizzate.
		 */
		~Lock();
		
		 /**
		 * Eseguito prima dell'utilizzo della risorsa effettua le seguenti operazioni:
		 * - Verifica la disponibilit della risorsa, rimanendo in attesa sino
		 *   a quando essa sia disponibile.
		 * - Occupa la risorsa.
		 * 
		 * \param resource string che identifica la risorsa da occupare.
		 */
		void lock(const string &resource);

		/**
		 * Eseguito al termine dell'utilizzo di una risorsa effettua le seguenti operazioni:
		 * - Libera la risorsa.
		 * - Risveglia un thread in attesa di occupare la risorsa (con Lock::lock(string)).
		 * \param resource string che identifica la risorsa da liberare.
		 */
		void unlock(const string &resource);
	private:
		inline void alloc(const string &resource);
		inline void free(const string &resource);
		inline bool isfree(const string &resource);

		pthread_mutex_t mut;
		pthread_cond_t cond;
		set<string,fbfs::str_cmp> resources;
};

#endif
