/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream.h>
#include "NameServer.h"

/** \file NameServer.cc
 * Implementazione dei metodi della classe utils::NameServer
 * \ingroup Utils
 */
   
using namespace utils;

CORBA::Object_ptr NameServer::get_object( const CORBA::ORB_ptr orb, const utils::URI &uri) throw(Exception) {
	string str_contextname = uri.nameServerURL();
	string str_servicename = uri.serviceName();
	const char *contextname = str_contextname.c_str();
	const char *servicename = str_servicename.c_str();
	CosNaming::NamingContextExt_var nc;
	//cout << "Context:" << str_contextname << " Service:" << str_servicename << "\n";	
	try {
		CORBA::Object_var obj = orb->string_to_object( contextname );
		nc = CosNaming::NamingContextExt::_narrow(obj);
		if( CORBA::is_nil(nc) ) {
			cerr << "Failed to narrow to naming context";
			throw Exception();
		};
		CORBA::Object_ptr return_object = nc->resolve_str( servicename );
		return return_object;
	}
	catch(CosNaming::NamingContext::InvalidName &) {
		cout << "invalid name\n";
		throw InvalidNameException();
	}
	catch(CosNaming::NamingContext::NotFound &) {
		cout << "not found\n";
		throw NotFoundException();
	}
	catch(...) {
		cout << "other\n";
		throw Exception();
	};
};

CORBA::Boolean NameServer::unbindObject(const char* context, const char* name, const char* kind, const CORBA::ORB_ptr orb) {
	CosNaming::NamingContext_var rootContext;
	try {
		// Obtain a reference to the root context of the Name service:
		CORBA::Object_var obj;
		obj = orb->resolve_initial_references("NameService");

		// Narrow the reference returned.
		rootContext = CosNaming::NamingContext::_narrow(obj);
		if( CORBA::is_nil(rootContext) ) {
			cerr << "Failed to narrow the root naming context." << endl;
			return false;
		}
	}
	catch(CORBA::ORB::InvalidName& ex) {
		// This should not happen!
		cerr << "Service required is invalid [does not exist]." << endl;
		return false;
	}

	try {
		// Open the required context
		CosNaming::Name contextName;
		contextName.length(1);
		contextName[0].id   = (const char*) context; // string copied
		contextName[0].kind = (const char*) "utils"; // string copied

		CORBA::Object_var obj = rootContext->resolve(contextName);
		CosNaming::NamingContext_var testContext = CosNaming::NamingContext::_narrow(obj);
		if( CORBA::is_nil(testContext) ) {
			cerr << "Failed to narrow naming context." << endl;
			return false;
		}

		// Unbind the specified object
		CosNaming::Name objectName;
		objectName.length(1);
		objectName[0].id   = (const char*) name;
		objectName[0].kind = (const char*) kind;

		try {
			testContext->unbind(objectName);
		}
		catch(CosNaming::NamingContext::NotFound& ex) {
			cerr << "The object in not registered in the selected context." << endl;
		}
// It appears to bring omniNames in an inconsistent state.
//		try {
//			testContext->destroy();
//		}
//		catch(CosNaming::NamingContext::NotEmpty& ex) { }
	}
	catch(CORBA::COMM_FAILURE& ex) {
		cerr << "Caught system exception COMM_FAILURE -- unable to contact the "
			<< "naming service." << endl;
		return false;
	}
	catch(CORBA::SystemException&) {
		cerr << "Caught a CORBA::SystemException while using the naming service." << endl;
		return false;
	}
	return true;
}

CORBA::Boolean NameServer::bindObjectToName(const char* context, const char* name, const char* kind, const CORBA::ORB_ptr orb, const CORBA::Object_ptr objref)
{
	CosNaming::NamingContext_var rootContext;
	try {
		// Obtain a reference to the root context of the Name service:
		CORBA::Object_var obj;
		obj = orb->resolve_initial_references("NameService");

		// Narrow the reference returned.
		rootContext = CosNaming::NamingContext::_narrow(obj);
		if( CORBA::is_nil(rootContext) ) {
			cerr << "Failed to narrow the root naming context." << endl;
			return false;
		}
	}
	catch(CORBA::ORB::InvalidName& ex) {
		// This should not happen!
		cerr << "Service required is invalid [does not exist]." << endl;
		return false;
	}

	try {
		// Bind a context called "test" to the root context:
		CosNaming::Name contextName;
		contextName.length(1);
		contextName[0].id   = (const char*) context;
		contextName[0].kind = (const char*) PROJECT_NAME;

		CosNaming::NamingContext_var testContext;
		try {
			// Bind the context to root.
			testContext = rootContext->bind_new_context(contextName);
		}
		catch(CosNaming::NamingContext::AlreadyBound& ex) {
			// If the context already exists, this exception will be raised.
			// In this case, just resolve the name and assign testContext
			// to the object returned:
			CORBA::Object_var obj;
			obj = rootContext->resolve(contextName);
			testContext = CosNaming::NamingContext::_narrow(obj);
			if( CORBA::is_nil(testContext) ) {
				cerr << "Failed to narrow naming context." << endl;
				return false;
			}
		}

		// Bind objref with name Echo to the testContext:
		CosNaming::Name objectName;
		objectName.length(1);
		objectName[0].id   = (const char*) name;
		objectName[0].kind = (const char*) kind;

		try {
			testContext->bind(objectName, objref);
		}
		catch(CosNaming::NamingContext::AlreadyBound& ex) {
			testContext->rebind(objectName, objref);
		}
		// Note: Using rebind() will overwrite any Object previously bound
		//       to /test/Echo with obj.
		//       Alternatively, bind() can be used, which will raise a
		//       CosNaming::NamingContext::AlreadyBound exception if the name
		//       supplied is already bound to an object.

		// Amendment: When using OrbixNames, it is necessary to first try bind
		// and then rebind, as rebind on it's own will throw a NotFoundexception if
		// the Name has not already been bound. [This is incorrect behaviour -
		// it should just bind].
	}
	catch(CORBA::COMM_FAILURE& ex) {
		cerr << "Caught system exception COMM_FAILURE -- unable to contact the "
			<< "naming service." << endl;
		return false;
	}
	catch(CORBA::SystemException&) {
		cerr << "Caught a CORBA::SystemException while using the naming service." << endl;
		return false;
	}
	return true;
}
