/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _utils_NAMESERVER
#define _utils_NAMESERVER 1

/** \file NameServer.h
 * Definizione della classe utils::NameServer
 * \ingroup Utils
 */

#define PROJECT_NAME "fbfs"
#include <omniORB3/CORBA.h>

#include "URI.h"

namespace utils {
	class NameServer;
};

/** Collezione di metodi per un utilizzo conveniente del NameServer CORBA.
 * Questa classe fornisce una collezione di routine adatte all'espletamento di
 * compiti ricorrenti riguardanti l'uso di un NameServer CORBA.
 * In particolare sono forniti metodi per:
 * - Registrare un oggetto CORBA presso il NameServer associato ad un ORB.
 * - Deregistrare un oggetto CORBA precedentemente registrato.
 * - Ottenere un riferimento ad un oggetto CORBA registrato presso un NameServer,
 *   conoscendone l'URI (come definito nella classe utils::URI).
 *
 * La classe fornisce informazioni sugli eventi anomali tramite l'uso consistente
 * di eccezioni.
 * \ingroup Utils
 */
class utils::NameServer {
	public:
		/** Classe base da cui ereditano tutte le eccezioni emesse da metodi della
		 * classe utils::NameServer.
		 */
		class Exception {};

		/** Il metodo non  andato a buon fine per l'impossibilit di trovare qualcosa.
		 */
		class NotFoundException : public Exception {};

		/** Il mentodo non  andato a buon fine poich un nome non  valido.
		 */
		class InvalidNameException : public Exception {};
		
		/** Registra un oggetto CORBA presso un NameServer.
		 * Consente di registrare un oggetto CORBA presso il nameserver associato
		 * all'ORB orb, permettendo di specificare il tipo del servizio da registrare,
		 * il nome con cui si potr riferire il server in questione e il contesto di
		 * NameServer presso cui  registrato il servizio.
		 * \param orb ORB che fornisce il NameServer per la registrazione dell'oggetto.
		 * \param obj Oggetto CORBA di cui si richiede la registrazione.
		 * \param context Nome del contesto in cui registrare il servizio.
		 * \param name Nome del server attraverso il quale si potr accedere al servizio.
		 * \param kind Tipo di servizio fornito.
		 * \return true se l'operazione  andata a buon fine.
		 */
		static CORBA::Boolean
			bindObjectToName(const char *context, const char* name, const char* kind,
					const CORBA::ORB_ptr orb, const CORBA::Object_ptr obj);

		/** Deregistra un oggetto CORBA predecentemente registrato presso un NameServer.
		 * Consente di rimuovere un'entry nel NameServer CORBA associato ad un particolare ORB.
		 * \param orb ORB che fornisce il NameServer in cui  registrato il servizio da deregistrare.
		 * \param context Nome del contesto in cui  registrato il servizio.
		 * \param name Nome del server attraverso il quale si accede al servizio.
		 * \param kind Tipo del servizio fornito.
		 * \return true se l'operazione  andata a buon fine.
		 */
		static CORBA::Boolean
			unbindObject(const char *context, const char* name, const char* kind, const CORBA::ORB_ptr orb);

		/** Restituisce un riferimento ad un oggetto CORBA datone il suo URI.
		 * Restituisce un riferimento all'oggetto CORBA, demandandone la gestione ad un
		 * ORB desiderato.
		 * \param orb ORB che si occupa della gestione dell'oggetto restituito
		 * \param uri URI del servizio che si desidera referenziare.
		 * \return riferimento all'oggetto referenziato
		 * \exception NameServer::InvalidNameException l'URI non  valido.
		 * \exception NameServer::NotFoundException presso l'URI desiderato non  presente alcun servizio.
		 * \exception NameServer::Exception l'operazione non  andata a buon fine per altri motivi.
		 */
		static CORBA::Object_ptr
			get_object( const CORBA::ORB_ptr orb, const utils::URI &uri ) throw(Exception);
};

#endif
