/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _utils_SHARELOCK
#define _utils_SHARELOCK

/** \file ShareLock.h
 * Definizione della classe utils::ShareLock.
 * \ingroup Utils
 */

#include <multiset.h>
#include <string>
#include "String.h"

namespace utils {
	class ShareLock;
};

/**
 * Consente l'accesso esclusivo o condiviso ad una risorsa ad un gruppo di thread.
 * Dato un insieme di risorse critiche identificate ogniuna con una string, consente
 * di garantire l'accesso ad ogniuna di esse in modo esclusivo o condiviso ad un singolo
 * thread di esecuzione.
 * La politica di assegnazione delle risorse  la seguente:
 * - Una risorsa pu essere acquisita come Shared da pi richiedenti, a patto che non sia
 *   gi stata acquisita in modo Exclusive da un'altro richiedente.
 * - Una risorsa pu essere acquisita in modo Exclusive da un solo richiedente. Inoltre
 *   essa non pu essere ottenuta se essa  gi stata acquisita in modo Shared da altri
 *   richiedenti.
 * \ingroup Utils
 */
class utils::ShareLock {
	public:
		/**
		 * Crea un oggetto di tipo Lock, allocando le risorse necessarie al suo utilizzo.
		 */
		ShareLock();

		/**
		 * Distrugge un oggetto di tipo Lock, liberando le risorse utilizzate.
		 */
		~ShareLock();
		
		 /**
		 * Eseguito prima dell'utilizzo della risorsa effettua le seguenti operazioni:
		 * - Verifica la disponibilit della risorsa, rimanendo in attesa sino
		 *   a quando essa sia disponibile.
		 * - Occupa la risorsa.
		 * 
		 * \param resource string che identifica la risorsa da occupare.
		 * \param exclusive true qualora si desideri un'accesso Exclusive alla risorsa.
		 */
		void lock(const string &resource, bool exclusive=false);

		/**
		 * Eseguito al termine dell'utilizzo di una risorsa effettua le seguenti operazioni:
		 * - Libera la risorsa.
		 * - Risveglia un thread in attesa di occupare la risorsa (con Lock::lock(string)).
		 * \param resource string che identifica la risorsa da liberare.
		 */
		void unlock(const string &resource);
	private:
		inline void alloc(const string &resource, bool ex);
		inline void free(const string &resource);
		inline bool isbusy(const string &resource, bool ex);

		pthread_mutex_t mut;
		pthread_cond_t cond;
		multiset<string, fbfs::str_cmp> excl;	// basterebbe un set, ma cos risparmio codice, poich
												// multiset gi devo usarlo.
		multiset<string, fbfs::str_cmp> shared;
};

#endif
