/*
 *    FBFS Distributed Object Repository
 *    Copyright (C) 2001 Francesco V. Bassi
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _utils_URI
#define _utils_URI 1

#include <string>

/** \file URI.h
 * Definizione della classe utils::URI.
 * \ingroup Utils
 */

namespace utils {
	class URI;
};

/**
 * Fornisce un approccio conveniente per la gestione degli URI del tipo utils://.
 * I metodi forniti da questa classe permetto un facile parsing di un URI del tipo \e utils://
 * e la sua conversione negli URI del NameServer (del tipo \e corbaloc://) e nel nome
 * strutturato del servizio.
 * \ingroup Utils
 */
class utils::URI {
	public:
		/**
		 * Crea un nuovo URI partendo da una rappresentazione esterna canonica del tipo
		 * utils://<host>:<port>/<service>@<context>/<server>.
		 *
		 * Dove:
		 *
		 * \e <host>:<port>  l'indirizzo del NameServer presso il quale  registrato
		 * il servizio desiderato
		 *
		 * \e <service> indica il tipo di servizio desiderato
		 *
		 * \e <context> indica il contesto nel quale  registrato il server
		 *
		 * \e <server> indica il nome del server desiderato
		 *
		 * \param stringified_URI string con la rappresentazione esterna "canonica" dell'URI.
		 */
		URI(const string &stringified_URI);

		/**
		 * Crea un nuovo URI da una descrizione esplicita delle sue componenti.
		 *
		 * \param host hostname del server presso il quale  presente il NameServer CORBA sul
		 * 		quale  registrato il servizio richiesto.
		 * \param port porta su cui ascolta il suddetto NameServer.
		 * \param service stringa che identifica il tipo di servizio richiesto.
		 * \param context contesto nel quale  registrato il servizio richiesto.
		 * \param name nome del server che fornisce il servizio richiesto.
		 */
		URI(const string &host, const string &port, const string &service, const string &context, const string &name) :
			_isValid(true), _host(host), _port(port), _service(service),
			_context(context), _name(name) {};

		/**
		 * Restituisce la porta su cui ascolta il NameServer per il servizio richiesto.
		 * \return porta su cui ascolta il NameServer per il servizio richiesto.
		 */
		string port() const { return _port; };

		/**
		 * Restituisce l'hostname del server sul quale resiede il NameServer del servizio richiesto.
		 * \return hostname del server sul quale resiede il NameServer del servizio richiesto.
		 */
		string host() const { return _host; };

		/**
		 * Restituisce una string che identifica il tipo del servizio richiesto.
		 * \return string che identifica il tipo del servizio richiesto.
		 */
		string service() const { return _service; };

		/**
		 * Restituisce il contesto in cui  registrato il servizio richiesto.
		 * \return contesto in cui  registrato il servizio richiesto.
		 */
		string context() const { return _context; };

		/**
		 * Restituisce il nome del servizio richiesto.
		 * \return nome del servizio richiesto.
		 */
		string name() const { return _name; };

		/**
		 * Restituisce l'URI corbaloc del NameServer presso cui  registrato il servizio
		 * \return string che contiene una rappresentazione \e stringified del URI del NameServer
		 * associato al servizio richiesto. Consente di contattare il NameServer CORBA richiesto
		 * per poi procedere alla risoluzione del nome del servizio e la restituzione di un
		 * riferimento all'oggetto CORBA che rappresenta il servizio richiesto.
		 */
		string nameServerURL() const;

		/**
		 * Restituisce il nome completamente qualificato del servizio, all'interno del NameServer
		 * prescelto.
		 * \return string che contiene il nome in forma canonica del servizio richiesto. Essa 
		 * del tipo <context>.utils/<server>.<type>.
		 */
		string serviceName() const;

		/**
		 * Restituisce una string che rappresenta una rappresentazione esterna "canonica" dell'URI.
		 * \return string che rappresenta una rappresentazione esterna "canonica" dell'URI.
		 */
		string toString() const;

		/**
		 * Verifica la validit dell'URI.
		 * \return true se l'oggetto rappresenta un URI valido. Solo in questo caso gli altri
		 * metodi di questa classe restituiscono valori sensati.
		 * \return false questo venere di risultato pu essere prodotto dall'utilizzo del metodo
		 * URI:URI(string) con una stringa che non rappresenta un URI corretto.
		 */
		bool isValid() const { return _isValid; };
	private:
		bool _isValid;		//|< Flag di validit dell'URI
		string _host;		//!< hostname del NameServer
		string _port;		//!< port del NameServer
		string _service;	//!< nome del servizio richiesto
		string _context;	//!< contesto in cui  registrato il servizio richiesto
		string _name;		//!< nome del server che fornisce il servizio richiesto
};

#endif
